/*
 * Copyright (c) 2014 - Zephyr Software LLC
 *
 * This file may be used and modified for non-commercial purposes as long as
 * all copyright, permission, and nonwarranty notices are preserved.
 * Redistribution is prohibited without prior written consent from Zephyr
 * Software.
 *
 * Please contact the authors for restrictions applying to commercial use.
 *
 * THIS SOURCE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Zephyr Software
 * e-mail: jwd@zephyr-software.com
 * URL   : http://www.zephyr-software.com/
 *
 */

#ifdef DEBUG
#include "malloc.h"
#include "itox.h"
#include "strlen.h"
#include "null.h"

#else

#endif


#ifdef DEBUG
	#define print_str_debug print_str
	#define print_unsigned_long_long_debug print_unsigned_long_long
#else
	#define print_str_debug(a) 
	#define print_unsigned_long_long_debug(a)  
	#define assert(a)
#endif


#ifdef DEBUG
#define write_fd 2
void print_str(char *s)
{
	write(write_fd,s,strlen(s));
}
void print_unsigned_long_long(unsigned long long x)
{
	char buf[100];
	itox(x,buf);
	write(write_fd,buf,strlen(buf));
}
#endif

extern unsigned long long canaries_top_of_stack;
 
void verify_all_canaries(void) {
	void *canary_loc = NULL; 
	uint16_t offset = 0;
	uint64_t canary = 0;
	uint64_t old_canary = 0;
	uint16_t distance_to_first_canary = 0;
	/*
	 * Without the nop here, fix_canaries
	 * believes that this is a push canary operation.
	 */
	asm volatile ("mov %%fs:0x28, %0\n"
	              "nop\n"
	     : "+r" (old_canary)
			 :
			 :);
	/*
	 * A trick to learn nearly the value of RSP,
	 * without resorting to ASM.
	 */
	canary_loc = &canary_loc;

	/*
	 * Emulate the logic for calculating a new canary value
	 * to learn the value of the most recent canary.
	 */
	distance_to_first_canary=(old_canary&0xFFFF)-(((uint64_t)canary_loc)&0xFFFF);
	canary_loc = (void*)(((uint64_t)canary_loc) + distance_to_first_canary);
	/*
	 * Offset is really only the last 16 bits.
	 */
	offset = old_canary & 0xFFFF;

#ifdef DEBUG
	print_str_debug("initial canary distance: 0x");
	print_unsigned_long_long_debug(distance_to_first_canary);
	print_str_debug("\n");
	print_str_debug("fs:0x28:                 0x");
	print_unsigned_long_long_debug(old_canary);
	print_str_debug("\n");
	print_str_debug("offset:                    ");
	print_unsigned_long_long_debug(offset);
	print_str_debug("\n");
#endif
	/*
	 * The last 16 bits don't matter for
	 * the canary value.
	 */
	old_canary >>= 16;

	/*
	 * Walk the canary stack and
	 * update as we go.
	 */
	while ((uint64_t)canary_loc < canaries_top_of_stack) {
#if DEBUG
		print_str_debug("We are inside the rewriting loop.\n");
		print_str_debug("canary_loc:");
		print_unsigned_long_long_debug(canary_loc);
		print_str_debug("\n");
#endif
/*
		printf("*canary_loc: 0x%lx\n", *((uint64_t*)(canary_loc)));
*/
#if DEBUG
		print_str_debug("Check: ");
		print_unsigned_long_long_debug((*((uint64_t*)(canary_loc)))>>16);
		print_str_debug(" ?= ");
		print_unsigned_long_long_debug(old_canary);
		print_str_debug("\n");
#endif
		if (((*((uint64_t*)(canary_loc))) >> 16) != old_canary) {
			__stack_chk_fail();
			break;
		}
	
		canary_loc += *(uint64_t*)canary_loc & 0xFFFF;

#if DEBUG
		print_str_debug("canary_loc':");
		print_unsigned_long_long_debug(canary_loc);
		print_str_debug("\n");
#endif

		/*
		 * TODO: Defend against nefariousness: 
		 * 1. Attacker setting offset to 0 -> infinite loop
		 * 2. Really deep stacks?
		 * 3. etc?
		 */
	}
	print_str_debug("successfully done rewriting all canary values.\n");
}

/*
 */
void zipr_debug_canary_callback(unsigned long long fs0x28,
	unsigned long long reg, unsigned long long rsp)
{
#ifdef DEBUG
	print_str_debug("In zipr_debug_canary_callback,\n");
	print_str_debug("fs:0x28=");
	print_unsigned_long_long_debug(fs0x28);
	print_str_debug("\n");
	print_str_debug("reg    =");
	print_unsigned_long_long_debug(reg);
	print_str_debug("\n");
	print_str_debug("rsp    =");
	print_unsigned_long_long_debug(rsp);
	print_str_debug("\n");
	verify_all_canaries();
#endif
}
