#include <stdint.h>
#include <syscall.h>
#include <null.h>

#include "infer.h"
#include "inferutil.h"

void print_str2(char *s)
{
	write(1,s,strlen(s));
}

static void send_response(const int fd, struct response *res)
{
	write(fd, res, sizeof(struct response));
}

static void send_error_response(const int fd)
{
        struct response res;
        clear_response(&res);
        send_response(fd, &res);
}

static void handleAlloc(struct request *req, struct response *res)
{
	int i;
	int size = req->arg1.val.num;
	int value = req->arg2.val.num;
	char *newmem = NULL; 

	// allocate memory
	int retval = cgc_allocate(size, 0, &newmem);

	if (retval == 0)
	{
		// initialize to specified value
		for (i = 0; i < size; ++i)
			newmem[i] = value;

		res->ok = 1;
		res->outarg.type = ARG_PTR;
		res->outarg.val.addr = newmem;
	}
	else
	{
		res->ok = 0;
	}
}

/* <bytes> READ <addr> <numBytes> */
static void handleRead(struct request *req, struct response *res)
{
        uintptr_t ptr = req->arg1.val.addr;
        int num_bytes = req->arg2.val.num;
        int i;

        if (num_bytes <= MAX_NUM_BYTES)
                res->ok = 1;
        else
        {
                num_bytes = MAX_NUM_BYTES;
                res->ok = 0;
        }

        res->outarg.type = ARG_BYTES;
        res->outarg.val.bytes.num_bytes = num_bytes;

        for (i = 0; i < num_bytes; ++i)
        {
                char *tmp = ptr;
                res->outarg.val.bytes.bytes[i] = tmp[i];
        }
}

/* WRITE <addr> <bytes> */
static void handleWrite(struct request *req, struct response *res)
{
        uintptr_t ptr = req->arg1.val.addr;
        int num_bytes = req->arg2.val.bytes.num_bytes;
        char *bytes = req->arg2.val.bytes.bytes;
        int i;

//        printf("WRITE %p %d\n", ptr, num_bytes);

/*
	char buf[2048];
	print_str2("handleWrite: #bytes: ");
	itox(num_bytes, buf);
	print_str2(buf);
	print_str2("\nhandleWrite: address: ");
	itox(ptr, buf);
	print_str2(buf);
	print_str2("\n");
*/

//	print_str2("handleWrite: [");
        for (i = 0; i < num_bytes; ++i)
	{
                ((char*)ptr)[i] = bytes[i];
//		print_str2(&bytes[i]);
	}
//	print_str2("]\n");

	// read back for grins
/*
	print_str2("read back newly allocated pointer: ");
	print_str(ptr);
	print_str2("\n");
*/

	// assume strlen is at 08049678
/*
        int (*fn)(char*);
	fn = 0x8049678;
	print_str2("handleWrite: size = ");
	int size = (*fn)((char*)ptr);
	itox(ptr, buf);
	print_str2(buf);
	print_str2("\n");
	
*/
        res->ok = 1;
}

static uintptr_t get_func_arg(struct argument *arg)
{
        if (arg->type == ARG_INT)
                return (uintptr_t) arg->val.num;
        else if (arg->type == ARG_PTR)
                return (uintptr_t) arg->val.addr;
        else if (arg->type == ARG_BYTES)
                return (uintptr_t) arg->val.bytes.bytes;
        else
                return (uintptr_t) 0x0;
}

/* outarg CALL <fn*> [arg] [arg] [arg] */
static void handleCall(struct request *req, struct response *res)
{
	char buf[2048];
        struct argument outarg; // void, int, void*

        struct argument *arg1 = &req->arg1;   // int, void*
        struct argument *arg2 = &req->arg2;   // int, void*
        struct argument *arg3 = &req->arg3;   // int, void*
        struct argument *arg4 = &req->arg4;   // int, void*

        uintptr_t retval;
        uintptr_t (*fn)(uintptr_t, ...);  
        uintptr_t fa1, fa2, fa3;

        fn = arg1->val.addr;

        fa1 = get_func_arg(arg2); // arg2 of request maps to function argument 1
        fa2 = get_func_arg(arg3); // arg3 of request maps to function argument 2
        fa3 = get_func_arg(arg4); // arg4 of request maps to function argument 3

			print_str2("arg values:");
			print_str2("arg1: ");
			itox(fa1, buf);
			print_str2(buf);
			print_str2(" arg2: ");
			itox(fa2, buf);
			print_str2(buf);
			print_str2(" arg3: ");
			itox(fa3, buf);
			print_str2(buf);

			print_str2("\n");

//        printf("CALL fn:%p #args: %d ret_type: %d\n", fn, req->num_args, req->outarg_type);

        if (req->num_args >= 1 && req->num_args <= 4)
        {
		print_str2("handleCall(): call backend function \n");

		retval = (*fn)(fa1, fa2, fa3);

/*
        	if (req->num_args <= 2)
	                retval = (*fn)((void*)fa1);
        	else if (req->num_args == 3)
	                retval = (*fn)((void*)fa1, fa2);
        	else if (req->num_args == 4)
	                retval = (*fn)((void*)fa1, fa2, fa3);
*/

		print_str2("handleCall(): done calling backend function \n");

                if (req->outarg_type == ARG_INT) {
			print_str2("handleCall(): expected return type: int\n");
                        res->ok = 1;
                        res->outarg.type = ARG_INT;
                        res->outarg.val.num = retval;
                        return;
                } else if (req->outarg_type == ARG_PTR) {
                        res->ok = 1;
                        res->outarg.type = ARG_PTR;
                        res->outarg.val.addr = retval;
			print_str2("handleCall(): expected return type: ptr\n");
                        return;
                } else if (req->outarg_type == ARG_NONE) {
                        res->ok = 1;
                        res->outarg.type = ARG_NONE;
                        res->outarg.val.num = retval;
			print_str2("handleCall(): expected return type: none\n");
		}
		else
			print_str2("handleCall(): error: unexpected outarg type\n");
        }
	else
		print_str2("handleCall(): error: unexpected #args\n");
}


static int handleCommand(const int fd, struct request *req, struct response *res)
{
	char buf[1024];
	print_str2("handleCommand(): enter\n");
	clear_response(res);

	switch(req->command) {
		case CMD_ALLOC:
			print_str2("handleCommand(): CMD_ALLOC: ");
			print_str2(" size:");
			itox(req->arg1.val.num, buf);
			print_str2(buf);
			print_str2(" value:");
			itox(req->arg2.val.num, buf);
			print_str2(buf);
			print_str2("\n");

			handleAlloc(req, res);

			print_str2("ALLOC address:");
			itox(res->outarg.val.addr, buf);
			print_str2(buf);
			print_str2("\n");


			send_response(fd, res);
			return 0;
			break;
		case CMD_CALL:
			itox(req->arg1.val.num, buf);
			print_str2("handleCommand(): CMD_CALL arg1: ");
			print_str2(buf);

			itox(req->arg2.val.num, buf);
			print_str2(" xarg2: ");
			print_str2(buf);

			itox(req->arg3.val.num, buf);
			print_str2(" xarg3: ");
			print_str2(buf);

			itox(req->arg4.val.num, buf);
			print_str2(" xarg4: ");
			print_str2(buf);

			print_str2(" return_type: ");
			itox(req->outarg_type, buf);
			print_str2(buf);

			print_str2("\n");

			handleCall(req, res);
			send_response(fd, res);
			return 0;
			break;
		case CMD_READ:
			print_str2("handleCommand(): CMD_READ\n");
			handleRead(req, res);
			send_response(fd, res);
			return 0;
			break;
		case CMD_WRITE:
			print_str2("handleCommand(): CMD_WRITE\n");
			handleWrite(req, res);
			send_response(fd, res);
			return 0;
   			break;
		case CMD_QUIT:
			print_str2("handleCommand(): CMD_QUIT\n");
			return 1;
			break;
		case CMD_NONE:
			print_str2("handleCommand(): CMD_NONE\n");
			return 0;
			break;
		default:
			print_str2("handleCommand(): unknown command\n");
			return 0;
			break;
	}

	print_str2("handleCommand(): exit\n");
	return 1;
}

void commandLoop()
{
	int fdin = CINDERELLA_READ;
	int fdout = CINDERELLA_WRITE;
	int done = 0;
	int bytes_read;
	char buf[sizeof(struct request) + 1024];

	struct request req;
	struct response res;

	print_str2("commandLoop(): enter here\n");

	char s[1024];

	do {
		clear_request(&req); 
		int bytes_read = read(fdin, &req, sizeof(struct request));
		itox(bytes_read, s);
		
		print_str2("commandLoop(): ... bytes read: \n");
		print_str2(s);
		print_str2("\n");

		done = handleCommand(fdout, &req, &res);
	} while (!done);

	print_str2("commandLoop(): exit\n");
}
